from __future__ import annotations
from ...ballot import RankBallot
from .csv_utils import (
    _validate_csv_ballot_weight,
    _validate_csv_ballot_voter_set,
    _validate_csv_ballot_row_break_idxs,
)
from typing import Tuple


def _parse_profile_data_from_rank_csv(
    csv_data: list[list[str]],
) -> Tuple[dict[str, str], int, bool, list[int]]:
    """
    Parse the profile data from a PreferenceProfile csv.

    Args:
        csv_data (list[list[str]]): Data from csv.

    Returns:
        Tuple[dict[str, str], int,  bool, list[int]]:
            inv_candidate_mapping, max_ranking_length,
            includes_voter_set, break_indices
    """
    candidate_row = [c_tuple.strip("()").split(":") for c_tuple in csv_data[2]]
    inv_candidate_mapping = {prefix: cand for cand, prefix in candidate_row}

    max_ranking_length = int(csv_data[4][0])

    includes_voter_set = csv_data[6][0] == "True"

    ballot_data_column_names = csv_data[8]
    break_indices = [
        i for i, col_name in enumerate(ballot_data_column_names) if col_name == "&"
    ]

    return (
        inv_candidate_mapping,
        max_ranking_length,
        includes_voter_set,
        break_indices,
    )


def _parse_ballot_from_rank_csv(
    ballot_row: list[str],
    includes_voter_set: bool,
    break_indices: list[int],
    inv_candidate_mapping: dict[str, str],
) -> RankBallot:
    """
    Parse a ballot from a PreferenceProfile csv row.

    Args:
        ballot_row (list[str]): Row from the csv file containing ballot data.
        includes_voter_set (bool): Whether or not the csv contains voter sets.
        break_indices (list[int]): Where the columns of the csv change from one data type to
            another.
        inv_candidate_mapping (dict[str, str]): The inverted candidate mapping of prefix
            to the cand.

    Returns:
        RankBallot: Ballot formatted from row of csv.
    """
    ranking_start = 0
    ranking_end = break_indices[0]

    formatted_ranking = None
    voter_set = set()

    try:
        weight = float(ballot_row[break_indices[0] + 1])
    except Exception:
        raise RuntimeError(
            f"Invalid weight format in ballot row: {ballot_row[break_indices[0] + 1]}. Weight "
            "must be float."
        )

    ranking = ballot_row[ranking_start:ranking_end]
    temp_ranking = [
        cand_set.strip("{}").split(", ") for cand_set in ranking if cand_set != ""
    ]
    formatted_ranking = tuple(
        [
            (
                frozenset(inv_candidate_mapping[c.strip("'")] for c in cand_set)
                if cand_set != [""]
                else frozenset()
            )
            for cand_set in temp_ranking
        ]
    )

    if includes_voter_set:
        voter_set = set(v.strip() for v in ballot_row[break_indices[-1] + 1 :])

    return RankBallot(ranking=formatted_ranking, voter_set=voter_set, weight=weight)


def _validate_rank_csv_header_values(header_data: list[list[str]]):
    """
    Validate that the values of the header rows are correct.

    Args:
        header_data (list[list[str]]): The rows of the csv file.

    Raises:
        ValueError: If csv header rows are improperly formatted for VoteKit.
    """
    boiler_plate = (
        "This usually indicates that you are loading a csv that was not made with "
        "PreferenceProfile.to_csv()."
    )

    if len(header_data[4]) != 1:
        raise ValueError(
            (
                "csv file is improperly formatted. Row 4 should be a single non-negative integer"
                " denoting the max ranking length, "
                f"not {header_data[4]}. " + boiler_plate
            )
        )

    try:
        max_ranking_length = int(header_data[4][0])

        if max_ranking_length < 0:
            raise ValueError(
                (
                    "csv file is improperly formatted. Row 4 should be a single"
                    " non-negative integer denoting the max ranking length, not "
                    f"{max_ranking_length}. " + boiler_plate
                )
            )
    except ValueError:
        raise ValueError(
            (
                "csv file is improperly formatted. Row 4 should be a single"
                " non-negative integer denoting the max ranking length, not "
                f"{header_data[4][0]}. " + boiler_plate
            )
        )

    if any(char not in c_tuple for c_tuple in header_data[2] for char in "(:)"):
        raise ValueError(
            (
                "csv file is improperly formatted. Row 2 should contain tuples mapping candidates "
                "to their unique prefixes. For example, (Chris:Ch), (Colleen: Co). "
                f"Not {header_data[2]}. " + boiler_plate
            )
        )

    if len(header_data[4]) != 1:
        raise ValueError(
            (
                "csv file is improperly formatted. Row 4 should be a single non-negative integer, "
                f"not {header_data[4]}. " + boiler_plate
            )
        )

    if len(header_data[6]) != 1 or header_data[6][0] not in ["True", "False"]:
        raise ValueError(
            (
                "csv file is improperly formatted. Row 6 should be 'True' or 'False'"
                " denoting the inclusion of a voter set,"
                f" not {header_data[6]}. " + boiler_plate
            )
        )


def _validate_rank_csv_header_rows(header_data: list[list[str]]):
    """
    Validate that the names of the header rows are correct.

    Args:
        header_data (list[list[str]]): The rows of the csv file.

    Raises:
        ValueError: If csv headers rows are improperly formatted for VoteKit.
    """
    boiler_plate = (
        "This usually indicates that you are loading a csv that was not made with "
        "PreferenceProfile.to_csv()."
    )

    if header_data[0] != ["VoteKit RankProfile"]:
        raise ValueError(
            (
                "csv file is improperly formatted. Row 0 should be 'VoteKit RankProfile',"
                f"not {header_data[0]}. " + boiler_plate
            )
        )

    if header_data[1] != ["Candidates"]:
        raise ValueError(
            (
                "csv file is improperly formatted. Row 1 should be 'Candidates', "
                f"not {header_data[1]}. " + boiler_plate
            )
        )

    if header_data[3] != ["Max Ranking Length"]:
        raise ValueError(
            (
                "csv file is improperly formatted. Row 3 should be 'Max Ranking Length', "
                f"not {header_data[3]}. " + boiler_plate
            )
        )

    if header_data[5] != ["Includes Voter Set"]:
        raise ValueError(
            (
                "csv file is improperly formatted. Row 5 should be 'Includes Voter Set', "
                f"not {header_data[5]}. " + boiler_plate
            )
        )

    if header_data[-1] != ["="] * 10:
        raise ValueError(
            (
                "csv file is improperly formatted. Row 7 should be '=,=,=,=,=,=,=,=,=,=', "
                f"not {header_data[-1]}. " + boiler_plate
            )
        )


def _validate_rank_csv_header(header_data: list[list[str]]):
    """
    Validate that the values of the header rows are correct.

    Args:
        header_data (list[list[str]]): The rows of the header of the csv file.

    Raises:
        ValueError: If csv header is improperly formatted for VoteKit.
    """

    _validate_rank_csv_header_rows(header_data)
    _validate_rank_csv_header_values(header_data)


def _validate_rank_csv_ballot_header_row(
    ballot_rows: list[list[str]],
    max_ranking_length: int,
    include_voter_set: bool,
):
    """
    Validate that the ballot header row is formatted correctly.

    Args:
        ballot_rows (list[list[str]]): All ballot rows.
        max_ranking_length (int): The max ranking length.
        include_voter_set (bool): Whether or not there is a voter set.

    Raises:
        ValueError: If the ballot header is improperly formatted for VoteKit.
    """
    boiler_plate = (
        "This usually indicates that you are loading a csv that was not made with "
        "PreferenceProfile.to_csv()."
    )

    header_row = ballot_rows[0]
    ranking_cols = [f"Ranking_{i+1}" for i in range(max_ranking_length)]

    if max_ranking_length > 0 and any(
        r_col not in header_row for r_col in ranking_cols
    ):
        raise ValueError(
            (
                "csv file is improperly formatted. Row 8 should include 'Ranking_i' for "
                f"i going from 1 to max ranking length, not {header_row}. "
                + boiler_plate
            )
        )
    elif max_ranking_length == 0 and any("Ranking_" in col for col in header_row):
        raise ValueError(
            (
                "csv file is improperly formatted. Row 8 should not include 'Ranking_', "
                f"not {header_row}. " + boiler_plate
            )
        )

    if "Weight" not in header_row:
        raise ValueError(
            (
                "csv file is improperly formatted. Row 8 should include 'Weight' column. "
                + boiler_plate
            )
        )

    if (include_voter_set and "Voter Set" not in header_row) or (
        not include_voter_set and "Voter Set" in header_row
    ):
        raise ValueError(
            (
                "csv file is improperly formatted. Includes Voter Set is not set to the correct"
                f" value given the columns in row 8: {header_row}. " + boiler_plate
            )
        )


def _validate_rank_csv_ballot_ranking(
    ballot_row: list[str],
    row_index: int,
    candidate_prefixes: tuple[str],
    max_ranking_length: int,
):
    """
    Validate that the ballot ranking is formatted correctly.

    Args:
        ballot_row (list[str]): A ballot row.
        row_index (int): The index of the row in the csv, 0-indexed.
        candidate_prefixes (tuple[str]): The tuple of candidate prefixes.
        max_ranking_length (int): The max ranking length.

    Raises:
        ValueError: If the ballot ranking is improperly formatted for VoteKit.
    """

    boiler_plate = (
        "This usually indicates that you are loading a csv that was not made with "
        "PreferenceProfile.to_csv()."
    )

    break_idxs = [i for i, string in enumerate(ballot_row) if string == "&"]
    if max_ranking_length == 0 and break_idxs[1] - break_idxs[0] > 1:
        raise ValueError(
            (
                f"csv file is improperly formatted. Ballot in row {row_index} has rankings "
                f"but it should not: {ballot_row[break_idxs[0]:break_idxs[1]]}. "
                + boiler_plate
            )
        )
    elif max_ranking_length == 0:
        return

    for ranking_set in ballot_row[: break_idxs[0]]:
        if ranking_set == "":
            continue

        if ranking_set[0] != "{" or ranking_set[-1] != "}":
            raise ValueError(
                (
                    f"csv file is improperly formatted. Ballot in row {row_index} has "
                    f"improperly formatted ranking set: {ranking_set}. " + boiler_plate
                )
            )

        if ranking_set in ["", "{}"]:
            continue

        cand_set = ranking_set.strip("{}").split(", ")
        for c in cand_set:
            if c.strip("'") not in candidate_prefixes:
                raise ValueError(
                    (
                        f"csv file is improperly formatted. Ballot in row {row_index} has "
                        f"undefined candidate prefix: {c}. " + boiler_plate
                    )
                )


def _validate_rank_csv_ballot_rows(csv_data: list[list[str]]):
    """
    Validate that each ballot row is formatted correctly.

    Args:
        csv_data (list[list[str]]): The full csv.

    Raises:
        ValueError: If a row of the csv is improperly formatted for VoteKit.
    """
    candidate_row = csv_data[2]
    max_ranking_row = csv_data[4]
    include_voter_set_row = csv_data[6]

    candidate_tuples = [c_tuple.strip("()").split(":") for c_tuple in candidate_row]
    candidates, candidate_prefixes = zip(*candidate_tuples)

    max_ranking_length = int(max_ranking_row[0])
    include_voter_set = include_voter_set_row[0] == "True"

    _validate_rank_csv_ballot_header_row(
        csv_data[8:], max_ranking_length, include_voter_set
    )
    error_rows = []
    for i, ballot_row in enumerate(csv_data[9:]):
        try:
            _validate_csv_ballot_row_break_idxs(ballot_row, i)
            _validate_rank_csv_ballot_ranking(
                ballot_row, i + 9, candidate_prefixes, max_ranking_length
            )
            _validate_csv_ballot_weight(ballot_row, i + 9)
            _validate_csv_ballot_voter_set(ballot_row, i + 9, include_voter_set)
        except ValueError as e:
            error_rows.append((i + 9, e))

        if len(error_rows) == 10:
            raise ValueError(
                (
                    f"There are errors on at least ten lines of this csv: {error_rows}."
                    " Did you try to load a csv that was edited by hand and not "
                    "created with the PreferenceProfile.to_csv() method?"
                )
            )
    if len(error_rows) > 0:
        raise ValueError(
            (
                f"There are errors on {len(error_rows)} lines of this csv: {error_rows}."
                " Did you try to load a csv that was edited by hand and not "
                "created with the PreferenceProfile.to_csv() method?"
            )
        )


def _validate_rank_csv_format(csv_data: list[list[str]]):
    """
    Validate that the csv is properly formatted for VoteKit.

    Args:
        csv_data (list[list[str]]): The rows of the csv file.

    Raises:
        ValueError: If csv is improperly formatted for VoteKit.
    """
    _validate_rank_csv_header(csv_data[:8])
    _validate_rank_csv_ballot_rows(csv_data)



from .profile_bar_plot import (
    profile_bar_plot,
    profile_borda_plot,
    profile_mentions_plot,
    profile_fpv_plot,
    profile_ballot_lengths_plot,
)

from .multi_profile_bar_plot import (
    multi_profile_bar_plot,
    multi_profile_ballot_lengths_plot,
    multi_profile_borda_plot,
    multi_profile_fpv_plot,
    multi_profile_mentions_plot,
)


__all__ = [
    "profile_bar_plot",
    "profile_borda_plot",
    "profile_mentions_plot",
    "profile_fpv_plot",
    "profile_ballot_lengths_plot",
    "multi_profile_bar_plot",
    "multi_profile_ballot_lengths_plot",
    "multi_profile_borda_plot",
    "multi_profile_fpv_plot",
    "multi_profile_mentions_plot",
]



from .ranking import (
    RankingElection,
    Plurality,
    SNTV,
    Borda,
    STV,
    FastSTV,
    IRV,
    Alaska,
    DominatingSets,
    SequentialRCV,
    CondoBorda,
    TopTwo,
    PluralityVeto,
    RandomDictator,
    BoostedRandomDictator,
    RankedPairs,
)


from .scores import GeneralRating, Rating, Limited, Cumulative
from .approval import Approval, BlocPlurality

__all__ = [
    "RankingElection",
    "Plurality",
    "SNTV",
    "Borda",
    "STV",
    "FastSTV",
    "IRV",
    "SequentialRCV",
    "Alaska",
    "DominatingSets",
    "CondoBorda",
    "TopTwo",
    "GeneralRating",
    "Rating",
    "Limited",
    "Cumulative",
    "Approval",
    "BlocPlurality",
    "PluralityVeto",
    "RandomDictator",
    "BoostedRandomDictator",
    "RankedPairs",
]



from votekit.pref_profile import RankProfile
from votekit.matrices.candidate.comentions import comention
import numpy as np
from typing import Tuple
from votekit.matrices._utils import _convert_dict_to_matrix


def boost_prob(i: str, j: str, pref_profile: RankProfile) -> Tuple[float, float]:
    """
    Takes candidates i,j and a preference profile and computes the conditional
    P(mention i | mention j) and P(mention i). If candidate j is never mentioned,
    or if the profile has 0 total ballot weight, it returns
    numpy.nan for the respective probability.

    Args:
      i (str): Candidate.
      j (str): Candidate.
      pref_profile (RankProfile): Profile.

    Returns:
      tuple[float, float]: P(mention i | mention j), P(mention i)
    """

    i_mentions = 0.0
    j_mentions = 0.0
    both_mentions = 0.0

    for ballot in pref_profile.ballots:
        if comention(i, ballot):
            i_mentions += ballot.weight

        if comention(j, ballot):
            j_mentions += ballot.weight

        if comention([i, j], ballot):
            both_mentions += ballot.weight

    return (
        float(both_mentions) / j_mentions if (j_mentions != 0) else np.nan,
        (
            float(i_mentions) / pref_profile.total_ballot_wt
            if (pref_profile.total_ballot_wt != 0)
            else np.nan
        ),
    )


def boost_matrix(pref_profile: RankProfile, candidates: list[str]) -> np.ndarray:
    """
    Takes a profile and converts to a matrix
    where the i,j entry shows P(mention i | mention j) - P(mention i).
    Thus, the i,j entry shows the boost given to i by j.
    Computations use ballot weight. Non-symmetric matrix.
    Undefined entries are denoted with numpy.nan values.

    Args:
      pref_profile (RankProfile): Profile.
      candidates (list[str]): List of candidates to use. Indexing of this list matches indexing of
          output array.

    Returns:
        np.ndarray: Numpy array of boosts.
    """
    boost_matrix = {c: {c: 0.0 for c in candidates} for c in candidates}

    for i in candidates:
        for j in candidates:
            if i != j:
                cond, uncond = boost_prob(i, j, pref_profile)
                boost_matrix[i][j] = cond - uncond
            else:
                boost_matrix[i][j] = np.nan

    return _convert_dict_to_matrix(boost_matrix)



"""
Generate scored preference profiles using the name-Cumulative model.

The main API functions in this module are:

- `name_cumulative_profile_generator`: Generates a single preference profile using the name-Cumulative
    model.
- `name_cumulative_ballot_generator_by_bloc`: Generates preference profiles by bloc using the
    name-Cumulative model.
"""

import numpy as np
import apportionment.methods as apportion
from typing import Optional

from votekit.ballot import ScoreBallot
from votekit.pref_profile import ScoreProfile
from votekit.ballot_generator.bloc_slate_generator.model import BlocSlateConfig

# ===========================================================
# ================= Interior Work Functions =================
# ===========================================================


def _inner_name_cumulative(
    config: BlocSlateConfig, total_points: int
) -> dict[str, ScoreProfile]:
    """
    Inner function to generate cumulative profiles by bloc using the name-Cumulative model.

    Args:
        config (BlocSlateConfig): Configuration object containing all necessary parameters for
            working with a bloc-slate ballot generator.
        total_points (int): The total number of points to distribute among candidates.

    Returns:
        dict[str, ScoreProfile]: A dictionary whose keys are bloc strings and values are
            `ScoreProfile` objects representing the generated ballots for each bloc.
    """
    bloc_lst = config.blocs
    n_voters = int(config.n_voters)

    bloc_counts = apportion.compute(
        "huntington", list(config.bloc_proportions.values()), n_voters
    )
    if not isinstance(bloc_counts, list):
        bloc_counts = [int(bloc_counts)]

    ballots_per_bloc = dict(zip(bloc_lst, bloc_counts))
    pp_by_bloc: dict[str, ScoreProfile] = {}

    pref_by_bloc = config.get_combined_preference_intervals_by_bloc()
    rng = np.random.default_rng()

    for bloc in bloc_lst:
        num_ballots = int(ballots_per_bloc.get(bloc, 0))
        if num_ballots <= 0:
            pp_by_bloc[bloc] = ScoreProfile()
            continue

        pref = pref_by_bloc[bloc]
        non_zero_cands = list(pref.non_zero_cands)
        if not non_zero_cands:
            pp_by_bloc[bloc] = ScoreProfile()
            continue

        # config.get_combined_preference_intervals_by_bloc() should ensure normalization
        # for the non-zero candidates
        p = np.array([pref.interval[c] for c in non_zero_cands], dtype=float)
        assert abs(p.sum() - 1.0) < 1e-10, "PreferenceInterval not normalized"

        # Vectorized: one multinomial per ballot -> shape (num_ballots, n_cands)
        # Each row sums to n_voters and the entries are counts for each candidate
        counts = rng.multinomial(n=total_points, pvals=p, size=num_ballots)

        ballots = [
            ScoreBallot(scores=dict(zip(non_zero_cands, row.astype(float))), weight=1)
            for row in counts
        ]
        pp_by_bloc[bloc] = ScoreProfile(ballots=tuple(ballots))

    return pp_by_bloc


# =================================================
# ================= API Functions =================
# =================================================


def name_cumulative_profile_generator(
    config: BlocSlateConfig,
    *,
    total_points: Optional[int] = None,
    group_ballots: bool = True,
) -> ScoreProfile:
    """
    Generates a ScoreProfile using the name-Cumulative.

    This model samples with replacement from a combined preference interval and counts candidates
    with multiplicity.

    Args:
        config (BlocSlateConfig): Configuration object containing all necessary parameters for
            working with a bloc-slate ballot generator.

    Kwargs:
        total_points (Optional[int]): The total number of points to distribute among candidates.
            If None, defaults to the number of candidates in the configuration. Defaults to None.
        group_ballots (bool): If True, groups identical ballots in the resulting profile.
            Defaults to True.

    Returns:
        ScoreProfile: A `ScoreProfile` object representing the generated ballots.
    """
    config.is_valid(raise_errors=True)

    if total_points is None:
        total_points = len(config.candidates)
    if total_points <= 0:
        raise ValueError("total_points must be a positive integer")

    pp_by_bloc = _inner_name_cumulative(config, total_points=total_points)

    pp = ScoreProfile()
    for profile in pp_by_bloc.values():
        pp += profile

    if group_ballots:
        pp = pp.group_ballots()

    return pp


def name_cumulative_ballot_generator_by_bloc(
    config: BlocSlateConfig,
    *,
    total_points: Optional[int] = None,
    group_ballots: bool = True,
) -> dict[str, ScoreProfile]:
    """
    Generates a dictionary mapping bloc names to ScoreProfiles using the name-Cumulative model.

    This model samples with replacement from a combined preference interval and counts candidates
    with multiplicity.

    Args:
        config (BlocSlateConfig): Configuration object containing all necessary parameters for
            working with a bloc-slate ballot generator.

    Kwargs:
        total_points (Optional[int]): The total number of points to distribute among candidates.
            If None, defaults to the number of candidates in the configuration. Defaults to None.
        group_ballots (bool): If True, groups identical ballots in the resulting profile.
            Defaults to True.

    Returns:
        dict[str, ScoreProfile]: A dictionary whose keys are bloc strings and values are
            `ScoreProfile` objects representing the generated ballots for each bloc.
    """
    config.is_valid(raise_errors=True)

    if total_points is None:
        total_points = len(config.candidates)
    if total_points <= 0:
        raise ValueError("'total_points' must be a positive integer")

    pp_by_bloc = _inner_name_cumulative(config, total_points=total_points)

    if group_ballots:
        for bloc in pp_by_bloc:
            pp_by_bloc[bloc] = pp_by_bloc[bloc].group_ballots()

    return pp_by_bloc



from abc import ABC, abstractmethod
import networkx as nx  # type: ignore
from typing import Any


class Graph(ABC):
    """
    Base class for graph models, contains general graph algorithmics
    applicable to any implementation
    """

    def __init__(self, graph: nx.Graph):
        self.graph = graph
        self.node_weights: dict = {}  # store node weights to avoid acessing Nx.graph

    @abstractmethod
    def build_graph(self, *args: Any, **kwargs: Any) -> nx.Graph:
        pass

    @abstractmethod
    def draw(self, *args: Any, **kwags: Any):
        pass

    def distance_between_subsets(self, A: nx.Graph, B: nx.Graph):
        """Returns distance between A,B"""
        Gc = self.graph
        shortest_paths: list = []
        for a in A.nodes:
            shortest_paths.extend(nx.shortest_path_length(Gc, a, b) for b in B.nodes)
        return min(shortest_paths)

    def subgraph_neighborhood(self, center, radius: int = 2) -> nx.Graph:
        """Returns a ball around center of given radius in the graph of all ballots"""
        return nx.ego_graph(self.graph, center, radius)

    def k_heaviest_neighborhoods(self, k: int = 2, radius: int = 2):
        """Returns dict of k ball neighborhoods of
        given radius with their centers and weights
        """
        if self.node_weights == {} or sum(self.node_weights.values()) == 0:
            raise TypeError("no weights assigned to graph")

        cast_ballots = {x for x in self.node_weights.keys() if self.node_weights[x] > 0}

        max_balls = {}

        for _ in range(k):
            weight = 0
            if len(cast_ballots) == 0:
                break
            for center in cast_ballots:
                ball = self.subgraph_neighborhood(center, radius)
                relevant = cast_ballots.intersection(
                    set(ball.nodes)
                )  ##cast ballots inside the ball
                tmp = sum(self.node_weights[node] for node in relevant)
                if tmp > weight:
                    weight = tmp
                    max_center = center
                    max_ball = ball

            not_cast_in_max_ball = set(max_ball.nodes).difference(cast_ballots)
            max_ball.remove_nodes_from(not_cast_in_max_ball)
            max_balls[max_center] = (max_ball, weight)

            cast_ballots = cast_ballots.difference(set(max_ball.nodes))

        return max_balls



"""
Generate ranked preference profiles using the name-BradleyTerry model.

The main API functions in this module are:

- `name_bt_profile_generator`: Generates a single preference profile using the name-BradleyTerry
    model.
- `name_bt_profiles_by_bloc_generator`: Generates preference profiles by bloc using the
    name-BradleyTerry model.
- `name_bt_profile_generator_using_mcmc`: Generates a single preference profile using MCMC
    sampling from the name-BradleyTerry model.
- `name_bt_profiles_by_bloc_generator_using_mcmc`: Generates preference profiles by bloc using
    MCMC sampling from the name-BradleyTerry model.
"""

import itertools as it
import numpy as np
import random
from typing import Mapping, Optional
import apportionment.methods as apportion
import math
import pandas as pd
import sys

from votekit.ballot import RankBallot
from votekit.pref_profile import RankProfile
from votekit.ballot_generator.bloc_slate_generator.model import BlocSlateConfig
from votekit.ballot_generator.utils import system_memory

# ====================================================
# ================= Helper Functions =================
# ====================================================


def _calc_prob(permutations: list[tuple], cand_support_dict: dict) -> dict:
    """
    given a list of (possibly incomplete) rankings and the preference interval, \
    calculates the probability of observing each ranking

    Args:
        permutations (list[tuple]): a list of permuted rankings
        cand_support_dict (dict): a mapping from candidate to their \
        support (preference interval)

    Returns:
        dict: a mapping of the rankings to their probability
    """
    ranking_to_prob = {}
    for ranking in permutations:
        prob = 1
        for i in range(len(ranking)):
            cand_i = ranking[i]
            greater_cand_support = cand_support_dict[cand_i]
            for j in range(i + 1, len(ranking)):
                cand_j = ranking[j]
                cand_support = cand_support_dict[cand_j]
                prob *= greater_cand_support / (greater_cand_support + cand_support)
        ranking_to_prob[ranking] = prob
    return ranking_to_prob


def _make_bradley_terry_numerator(vals):
    """
    Given a list of values, returns the numerator of the Bradley-Terry probability
    for a given ranking.

    Args:
        vals (list): a list of values corresponding to the candidates in the ranking

    Returns:
        float: the numerator of the Bradley-Terry probability
    """
    ret = 1.0
    m = len(vals)

    # Faster than math.prod
    for i in range(m - 1):
        ret *= vals[i] ** (m - i - 1)
    return ret


def _bradley_terry_pdf(dct: Mapping[str, float]) -> dict[tuple[str, ...], float]:
    """
    Given a dictionary of candidates and their support, returns the probability density function
    over all possible rankings.

    Args:
        dct (Mapping[str, float]): a mapping from candidate to their support

    Returns:
        dict: a mapping of the rankings to their probability
    """
    weights = {
        perm: _make_bradley_terry_numerator([dct[i] for i in perm])
        for perm in it.permutations(dct.keys(), len(dct))
    }
    total_weight = sum(weights.values())

    return {k: v / total_weight for k, v in weights.items()}


def _check_name_bt_memory(config: BlocSlateConfig) -> None:
    """
    Check if there is enough memory to generate the profile using the name-BradleyTerry model.

    Args:
        config (BlocSlateConfig): Configuration object containing all necessary parameters for
            working with a bloc-slate ballot generator.

    Raises:
        ValueError: If there are more than 12 candidates.
        MemoryError: If there is not enough memory to generate the pmf.
    """
    n_cands = len(config.candidates)
    if n_cands > 12:
        raise ValueError(
            "The name-BradleyTerry model is not recommended for more than 12 candidates due to "
            "combinatorial explosion of generating the pmf."
        )

    mem = system_memory()
    pmf_size = math.factorial(n_cands)
    candidate_with_longest_name = max(config.candidates, key=len)
    est_bytes_pmf = pmf_size * sys.getsizeof(candidate_with_longest_name) * n_cands
    est_bytes_profile = (
        config.n_voters
        * n_cands
        * sys.getsizeof(frozenset({candidate_with_longest_name}))
    )
    est_bytes = float(est_bytes_pmf + est_bytes_profile)

    # fudge factor for overhead. Just tuned to a couple of machines, but gives pretty close
    # upper bound on memory usage while leaving room for other processes
    est_bytes *= 1.5
    if est_bytes > mem["available_gib"] * 2**30:
        raise MemoryError(
            f"Not enough memory to generate the profile. Estimated memory usage is "
            f"{est_bytes / 2**30:.1f} GiB, but only {mem['available_gib']:.1f} GiB is available."
        )


# ===========================================================
# ================= Interior Work Functions =================
# ===========================================================


def _inner_name_bradley_terry(config: BlocSlateConfig) -> dict[str, RankProfile]:
    """
    Sample from the BT distribution using direct sampling.

    This is an interior helper function that does the bulk of the work for generating
    name-BradleyTerry profiles. It is not recommended to use this function directly,
    as it does not perform any memory checks or grouping of ballots.

    Args:
        config (BlocSlateConfig): Configuration object containing all necessary parameters for
            working with a bloc-slate ballot generator.

    Returns:
        dict[str, RankProfile]: Generated preference profiles by bloc.
    """
    n_candidates = len(config.candidates)

    bloc_lst = config.blocs

    bloc_counts = apportion.compute(
        "huntington", list(config.bloc_proportions.values()), config.n_voters
    )
    if not isinstance(bloc_counts, list):
        if not isinstance(bloc_counts, int):
            raise TypeError(
                f"Unexpected type from apportionment got {type(bloc_counts)}"
            )

        bloc_counts = [bloc_counts]

    ballots_per_bloc = {bloc: bloc_counts[i] for i, bloc in enumerate(bloc_lst)}

    pp_by_bloc = {b: RankProfile() for b in bloc_lst}

    pref_interval_by_bloc_dict = config.get_combined_preference_intervals_by_bloc()

    for bloc in config.bloc_proportions.keys():
        n_ballots = ballots_per_bloc[bloc]

        # Directly initialize the list using good memory trick
        ballot_pool = np.full((n_ballots, n_candidates), frozenset("~"), dtype=object)
        zero_cands = pref_interval_by_bloc_dict[bloc].zero_cands
        single_bloc_pdf_dict = _bradley_terry_pdf(
            pref_interval_by_bloc_dict[bloc].interval
        )

        # Directly use the keys and values from the dictionary for sampling
        rankings, probs = zip(*single_bloc_pdf_dict.items())

        # The return of this will be a numpy array, so we don't need to make it into a list
        sampled_indices = np.array(
            np.random.choice(
                a=len(rankings),
                size=n_ballots,
                p=probs,
            ),
            ndmin=1,
        )

        for j, index in enumerate(sampled_indices):
            ranking = [frozenset({cand}) for cand in rankings[index]]

            # Add any zero candidates as ties only if they exist
            if zero_cands:
                ranking.append(frozenset(zero_cands))

            ballot_pool[j] = np.array(ranking)

        df = pd.DataFrame(ballot_pool)
        df.index.name = "Ballot Index"
        df.columns = [f"Ranking_{i + 1}" for i in range(n_candidates)]
        df["Weight"] = 1
        df["Voter Set"] = [frozenset()] * len(df)
        pp = RankProfile(
            candidates=config.candidates,
            df=df,
            max_ranking_length=n_candidates,
        )
        pp_by_bloc[bloc] = pp

    return pp_by_bloc


# TODO:
# - Improve the subsampling scheme here (memory + speed)
# - Construct ballots as tuples
# - Other speed improvements
def _bradley_terry_mcmc(
    n_ballots: int,
    pref_interval: Mapping[str, float],
    seed_ballot: RankBallot,
    zero_cands: Optional[frozenset[str]] = None,
    verbose: bool = False,
    burn_in_time: int = 0,
    chain_length: Optional[int] = None,
):
    """
    Sample from BT distribution for a given preference interval using MCMC. Defaults
    to continuous sampling and no burn-in time.

    Args:
        n_ballots (int): the number of ballots to sample
        pref_interval (Mapping[str, float]): the preference interval to determine BT distribution
        seed_ballot (RankBallot):  the seed ballot for the Markov chain
        zero_cands (frozenset[str], optional): candidates with zero support to be added as ties
        verbose (bool): If True, print the acceptance ratio of the chain. Defaults to False.
        burn_in_time (int): the number of ballots discarded in the beginning of the chain
        chain_length (Optional[int]): the length of the Markov Chain. Ballots are subsampled every
            chain_length//n_ballots steps from the chain until the desired number of ballots is
            reached. Defaults to None which sets the chain_length to the number of ballots in
            the config.
    """
    if zero_cands is None:
        zero_cands = frozenset()

    if chain_length is None:
        chain_length = n_ballots

    assert seed_ballot.ranking is not None
    # check that seed ballot has no ties
    for s in seed_ballot.ranking:
        if len(s) > 1:
            raise ValueError("Seed ballot contains ties")

    ballots = [RankBallot()] * n_ballots
    accept = 0
    current_ranking = list(seed_ballot.ranking)
    n_candidates = len(current_ranking)

    # presample swap indices
    burn_in_time = burn_in_time  # int(10e5)
    if verbose:
        print(f"Burn in time: {burn_in_time}")
    swap_indices = [
        (j1, j1 + 1)
        for j1 in random.choices(range(n_candidates - 1), k=n_ballots + burn_in_time)
    ]

    for i in range(burn_in_time):
        # choose adjacent pair to propose a swap
        j1, j2 = swap_indices[i]
        acceptance_prob = min(
            1,
            pref_interval[next(iter(current_ranking[j2]))]
            / pref_interval[next(iter(current_ranking[j1]))],
        )

        # if you accept, make the swap
        if random.random() < acceptance_prob:
            current_ranking[j1], current_ranking[j2] = (
                current_ranking[j2],
                current_ranking[j1],
            )
            accept += 1

    # generate MCMC sample
    for i in range(n_ballots):
        # choose adjacent pair to propose a swap
        j1, j2 = swap_indices[i]
        acceptance_prob = min(
            1,
            pref_interval[next(iter(current_ranking[j2]))]
            / pref_interval[next(iter(current_ranking[j1]))],
        )

        # if you accept, make the swap
        if random.random() < acceptance_prob:
            current_ranking[j1], current_ranking[j2] = (
                current_ranking[j2],
                current_ranking[j1],
            )
            accept += 1

        if len(zero_cands) > 0:
            ballots[i] = RankBallot(ranking=current_ranking + [zero_cands])
        else:
            ballots[i] = RankBallot(ranking=current_ranking)

    if verbose:
        print(
            f"Acceptance ratio as number accepted / total steps: "
            f"{accept / (n_ballots + burn_in_time):.2}"
        )

    if -1 in ballots:
        raise ValueError("Some element of ballots list is not a ballot.")

    if n_ballots > chain_length:
        raise ValueError(
            "The Markov Chain length cannot be less than the number of ballots."
        )

    if verbose:
        print(f"The number of ballots before is {len(ballots)}")

    # Subsample evenly ballots
    ballots = [
        ballots[i * chain_length // n_ballots + chain_length // (2 * n_ballots)]
        for i in range(n_ballots)
    ]

    if verbose:
        print(f"The number of ballots after is {len(ballots)}")

    pp = RankProfile(ballots=ballots)  # type: ignore
    return pp


def _inner_name_bradley_terry_mcmc(
    config: BlocSlateConfig,
    *,
    verbose: bool = False,
    burn_in_time: int = 0,
    chain_length: Optional[int] = None,
) -> dict[str, RankProfile]:
    """
    Sample from the BT distribution using Markov Chain Monte Carlo.

    Args:
        number_of_ballots (int): The number of ballots to generate.

    Kwargs:
        verbose (bool): If True, print the acceptance ratio of the chain. Defaults to False.
        burn_in_time (int): the number of ballots discarded in the beginning of the chain.
            Defaults to 0.
        chain_length (Optional[int]): the length of the Markov Chain. Ballots are subsampled every
            chain_length//n_ballots steps from the chain until the desired number of ballots is
            reached. Defaults to None which sets the chain_length to the number of ballots in
            the config.

    Returns:
        Union[RankProfile, Tuple]
    """
    bloc_lst = config.blocs

    bloc_counts = apportion.compute(
        "huntington", list(config.bloc_proportions.values()), config.n_voters
    )
    if not isinstance(bloc_counts, list):
        if not isinstance(bloc_counts, int):
            raise TypeError(
                f"Unexpected type from apportionment got {type(bloc_counts)}"
            )

        bloc_counts = [bloc_counts]

    ballots_per_bloc = {bloc: bloc_counts[i] for i, bloc in enumerate(bloc_lst)}

    pp_by_bloc = {b: RankProfile() for b in bloc_lst}
    pref_interval_by_bloc_dict = config.get_combined_preference_intervals_by_bloc()

    for bloc in bloc_lst:
        n_ballots = ballots_per_bloc[bloc]
        pref_interval = pref_interval_by_bloc_dict[bloc]
        pref_interval_dict = pref_interval.interval
        non_zero_cands = pref_interval.non_zero_cands
        zero_cands = pref_interval.zero_cands

        seed_ballot = RankBallot(
            ranking=tuple([frozenset({c}) for c in non_zero_cands])
        )
        pp = _bradley_terry_mcmc(
            n_ballots,
            pref_interval_dict,
            seed_ballot,
            zero_cands=zero_cands,
            verbose=verbose,
            burn_in_time=burn_in_time,
            chain_length=chain_length,
        )

        pp_by_bloc[bloc] = pp

    return pp_by_bloc


# =================================================
# ================= API Functions =================
# =================================================


def name_bt_profiles_by_bloc_generator(
    config: BlocSlateConfig, *, group_ballots=True
) -> dict[str, RankProfile]:
    """
    Generate preference profiles by bloc using the name-BradleyTerry model.

    The probability of sampling the ranking :math:`X>Y>Z` is proportional to
    :math:`P(X>Y)*P(X>Z)*P(Y>Z)`. These individual probabilities are based on the preference
    interval: :math: `P(X>Y) = x/(x+y)`.

    Args:
        config (BlocSlateConfig): Configuration object containing all necessary parameters for
            working with a bloc-slate ballot generator.
        group_ballots (bool): If True, group identical ballots in the returned profile and
            set the weight accordingly. Defaults to True.

    Returns:
        dict[str, RankProfile]: Generated preference profiles by bloc.
    """
    _check_name_bt_memory(config)
    config.is_valid(raise_errors=True)
    pp_by_bloc = _inner_name_bradley_terry(config)
    if group_ballots:
        for bloc in pp_by_bloc:
            pp_by_bloc[bloc] = pp_by_bloc[bloc].group_ballots()

    return pp_by_bloc


def name_bt_profile_generator(
    config: BlocSlateConfig, *, group_ballots=True
) -> RankProfile:
    """
    Generate a preference profile using the name-BradleyTerry model.

    The probability of sampling the ranking :math:`X>Y>Z` is proportional to
    :math:`P(X>Y)*P(X>Z)*P(Y>Z)`. These individual probabilities are based on the preference
    interval: :math: `P(X>Y) = x/(x+y)`.

    Args:
        config (BlocSlateConfig): Configuration object containing all necessary parameters for
            working with a bloc-slate ballot generator.
        group_ballots (bool): If True, group identical ballots in the returned profile and
            set the weight accordingly. Defaults to True.

    Returns:
        RankProfile: Generated preference profile.
    """
    _check_name_bt_memory(config)
    config.is_valid(raise_errors=True)
    pp_by_bloc = _inner_name_bradley_terry(config)

    # combine the profiles
    pp = RankProfile()
    for profile in pp_by_bloc.values():
        pp += profile

    if group_ballots:
        pp.group_ballots()
    return pp


def name_bt_profile_generator_using_mcmc(
    config: BlocSlateConfig,
    *,
    group_ballots=True,
    verbose: bool = False,
    burn_in_time: int = 0,
    chain_length: Optional[int] = None,
) -> RankProfile:
    """
    Generate a preference profile using MCMC sampling from the name-BradleyTerry model.

    Args:
        config (BlocSlateConfig): Configuration object containing all necessary parameters for
            working with a bloc-slate ballot generator.

    Kwargs:
        group_ballots (bool): If True, group identical ballots in the returned profile and
            set the weight accordingly. Defaults to True.
        verbose (bool): If True, print the acceptance ratio of the chain. Defaults to False.
        burn_in_time (int): the number of ballots discarded in the beginning of the chain.
            Defaults to 0.
        chain_length (Optional[int]): the length of the Markov Chain. Ballots are subsampled every
            chain_length//n_ballots steps from the chain until the desired number of ballots is
            reached. Defaults to None which sets the chain_length to the number of ballots in
            the config.

    Returns:
        RankProfile: Generated preference profile.
    """
    config.is_valid(raise_errors=True)
    pp_by_bloc = _inner_name_bradley_terry_mcmc(
        config, verbose=verbose, burn_in_time=burn_in_time, chain_length=chain_length
    )
    # combine the profiles
    pp = RankProfile()
    for profile in pp_by_bloc.values():
        pp += profile

    if group_ballots:
        pp.group_ballots()
    return pp


def name_bt_profiles_by_bloc_generator_using_mcmc(
    config: BlocSlateConfig,
    *,
    group_ballots=True,
    verbose: bool = False,
    burn_in_time: int = 0,
    chain_length: Optional[int] = None,
) -> dict[str, RankProfile]:
    """
    Generate a preference profile dictionary by bloc using MCMC sampling from the
    name-BradleyTerry model.

    Args:
        config (BlocSlateConfig): Configuration object containing all necessary parameters for
            working with a bloc-slate ballot generator.

    Kwargs:
        group_ballots (bool): If True, group identical ballots in the returned profile and
            set the weight accordingly. Defaults to True.
        verbose (bool): If True, print the acceptance ratio of the chain. Defaults to False.
        burn_in_time (int): the number of ballots discarded in the beginning of the chain.
            Defaults to 0.
        chain_length (Optional[int]): the length of the Markov Chain. Ballots are subsampled every
            chain_length//n_ballots steps from the chain until the desired number of ballots is
            reached. Defaults to None which sets the chain_length to the number of ballots in
            the config.

    Returns:
        dict[str, RankProfile]: Generated preference profiles by bloc.
    """
    config.is_valid(raise_errors=True)
    pp_by_bloc = _inner_name_bradley_terry_mcmc(
        config, verbose=verbose, burn_in_time=burn_in_time, chain_length=chain_length
    )

    if group_ballots:
        for bloc in pp_by_bloc:
            pp_by_bloc[bloc] = pp_by_bloc[bloc].group_ballots()

    return pp_by_bloc


# ===========================================================================
# ================= Additional Functions Not Yet Integrated =================
# ===========================================================================


def _bradley_terry_mcmc_shortcut(
    n_ballots,
    pref_interval,
    seed_ballot,
    zero_cands={},
    verbose=False,
    burn_in_time=0,
    chain_length=None,
    BURN_IN_TIME=100000,
):
    """
    Sample from BT using MCMC on the shortcut ballot graph

    Args:
        n_ballots (int): the number of ballots to sample
        pref_interval (dict): the preference interval to determine BT distribution
        sub_sample_length (int): how many attempts at swaps to make before saving ballot
        seed_ballot: Ballot, the seed ballot for the Markov chain
        burn_in_time (int): the number of ballots discarded in the beginning of the chain
        chain_length (Optional[int]): the length of the Markov Chain. Ballots are subsampled every
            chain_length//n_ballots steps from the chain until the desired number of ballots is
            reached. Defaults to None which sets the chain_length to the number of ballots in
            the config.
    """
    # NOTE: Most of this has been copied from `_bradley_terry_mcmc`
    # TODO: Abstract the overlapping steps into another helper
    # function, and just pass the indices / transition probability
    # function

    if chain_length is None:
        chain_length = n_ballots

    # check that seed ballot has no ties
    for s in seed_ballot.ranking:
        if len(s) > 1:
            raise ValueError("Seed ballot contains ties")

    ballots = [-1] * n_ballots
    accept = 0
    current_ranking = list(seed_ballot.ranking)
    n_candidates = len(current_ranking)

    if verbose:
        print("MCMC on shortcut")

    burn_in_time = burn_in_time
    if verbose:
        print(f"Burn in time: {burn_in_time}")

    # precompute all the swap indices
    swap_indices = [
        tuple(sorted(random.sample(range(n_candidates), 2)))
        for _ in range(n_ballots + burn_in_time)
    ]

    for i in range(burn_in_time):
        # choose adjacent pair to propose a swap
        j1, j2 = swap_indices[i]
        j1_rank = j1 + 1
        j2_rank = j2 + 1
        if j2_rank <= j1_rank:
            raise Exception("MCMC on Shortcut: invalid ranks found")

        acceptance_prob = min(
            1,
            (pref_interval[next(iter(current_ranking[j2]))] ** (j2_rank - j1_rank))
            / (pref_interval[next(iter(current_ranking[j1]))] ** (j2_rank - j1_rank)),
        )

        # if you accept, make the swap
        if random.random() < acceptance_prob:
            current_ranking[j1], current_ranking[j2] = (
                current_ranking[j2],
                current_ranking[j1],
            )
            accept += 1

    # generate MCMC sample
    for i in range(n_ballots):
        # choose adjacent pair to propose a swap
        j1, j2 = swap_indices[i]
        j1_rank = j1 + 1
        j2_rank = j2 + 1
        if j2_rank <= j1_rank:
            raise Exception("MCMC on Shortcut: invalid ranks found")

        acceptance_prob = min(
            1,
            (pref_interval[next(iter(current_ranking[j2]))] ** (j2_rank - j1_rank))
            / pref_interval[next(iter(current_ranking[j1]))] ** (j2_rank - j1_rank),
        )

        # if you accept, make the swap
        if random.random() < acceptance_prob:
            current_ranking[j1], current_ranking[j2] = (
                current_ranking[j2],
                current_ranking[j1],
            )
            accept += 1

        if len(zero_cands) > 0:
            ballots[i] = RankBallot(ranking=current_ranking + [zero_cands])
        else:
            ballots[i] = RankBallot(ranking=current_ranking)

    if verbose:
        print(
            f"Acceptance ratio as number accepted / total steps: "
            f"{accept / (n_ballots + BURN_IN_TIME):.2}"
        )

    if -1 in ballots:
        raise ValueError("Some element of ballots list is not a ballot.")

    if n_ballots > chain_length:
        raise ValueError(
            "The Markov Chain length cannot be less than the number of ballots."
        )

    if verbose:
        print(f"The number of ballots before is {len(ballots)}")

    # Subsample evenly ballots
    ballots = [
        ballots[i * chain_length // n_ballots + chain_length // (2 * n_ballots)]
        for i in range(n_ballots)
    ]

    if verbose:
        print(f"The number of ballots after is {len(ballots)}")

    pp = RankProfile(ballots=ballots)
    return pp



"""
Generate ranked preference profiles using the slate-Plackett-Luce model.

The main API functions in this module are:

- `slate_pl_profile_generator`: Generates a single preference profile using the slate-Plackett-Luce
    model.
- `slate_pl_profiles_by_bloc_generator`: Generates preference profiles by bloc using the
    slate-Plackett-Luce model.
"""

import itertools as it
import numpy as np
import pandas as pd
from typing import Mapping
import apportionment.methods as apportion

from votekit.pref_profile import RankProfile
from votekit.ballot_generator import (
    sample_cohesion_ballot_types,
)
from votekit.ballot_generator.bloc_slate_generator.model import BlocSlateConfig
from votekit.ballot_generator.bloc_slate_generator.slate_utils import (
    _make_cand_ordering_by_slate,
    _convert_ballot_type_to_ranking,
)

# ===========================================================
# ================= Interior Work Functions =================
# ===========================================================


def _inner_slate_plackett_luce(
    config: BlocSlateConfig,
) -> dict[str, RankProfile]:
    """
    Inner function to generate preference profiles by bloc using the slate-Plackett-Luce model.

    This model first samples a ballot type by flipping a cohesion parameter weighted coin.
    It then fills out the ballot type via sampling without replacement from the interval
    (i.e. according to the name-Plackett-Luce model).

    Args:
        config (BlocSlateConfig): Configuration object containing all necessary parameters for
            working with a bloc-slate ballot generator.
        ballot_length (Optional[int]): Number of votes allowed per ballot. If None, this is
            set to the total number of candidates in the configuration. Defaults to None.

    Returns:
        dict[str, RankProfile]: Dictionary whose keys are bloc strings and values are
            ``RankProfile`` objects.
    """
    n_candidates = len(config.candidates)
    bloc_lst = config.blocs

    bloc_counts = apportion.compute(
        "huntington", list(config.bloc_proportions.values()), config.n_voters
    )
    if not isinstance(bloc_counts, list):
        if not isinstance(bloc_counts, int):
            raise TypeError(
                f"Unexpected type from apportionment got {type(bloc_counts)}"
            )

        bloc_counts = [bloc_counts]

    ballots_per_bloc = {bloc: bloc_counts[i] for i, bloc in enumerate(bloc_lst)}

    pref_profile_by_bloc: Mapping[str, RankProfile] = {
        b: RankProfile() for b in bloc_lst
    }

    pref_profile_by_bloc = {}

    for bloc in bloc_lst:
        n_ballots = ballots_per_bloc[bloc]
        ballot_pool = np.full((n_ballots, n_candidates), frozenset("~"))
        pref_intervals_by_slate_dict = config.get_preference_intervals_for_bloc(bloc)
        zero_cands = set(
            it.chain(*[pi.zero_cands for pi in pref_intervals_by_slate_dict.values()])
        )

        slate_to_non_zero_candidates = {
            s: [c for c in c_list if c not in zero_cands]
            for s, c_list in config.slate_to_candidates.items()
        }

        ballot_types = sample_cohesion_ballot_types(
            slate_to_non_zero_candidates=slate_to_non_zero_candidates,
            num_ballots=n_ballots,
            cohesion_parameters_for_bloc=config.cohesion_df.loc[bloc].to_dict(),  # type: ignore
        )

        for j, bt in enumerate(ballot_types):
            cand_ordering_by_slate = _make_cand_ordering_by_slate(
                config, pref_intervals_by_slate_dict
            )
            ranking = _convert_ballot_type_to_ranking(
                ballot_type=bt, cand_ordering_by_slate=cand_ordering_by_slate
            )
            if ranking is None:
                raise RuntimeError(
                    "Unexpeceted None from internal function _convert_ballot_type_to_ranking "
                    "Likely caused by an empty ballot type."
                )

            if len(zero_cands) > 0:
                ranking.append(frozenset(zero_cands))
            ballot_pool[j] = np.array(ranking)

        df = pd.DataFrame(ballot_pool)
        df.index.name = "Ballot Index"
        df.columns = [f"Ranking_{i + 1}" for i in range(n_candidates)]
        df["Weight"] = 1
        df["Voter Set"] = [frozenset()] * len(df)
        pp = RankProfile(
            candidates=config.candidates,
            df=df,
            max_ranking_length=n_candidates,
        )
        pref_profile_by_bloc[bloc] = pp

    return pref_profile_by_bloc


# =================================================
# ================= API Functions =================
# =================================================


def slate_pl_profile_generator(
    config: BlocSlateConfig,
    *,
    group_ballots: bool = True,
) -> RankProfile:
    """
    Generates a merged preference profile using the slate-Plackett-Luce model.

    This model first samples a ballot type by flipping a cohesion parameter weighted coin.
    It then fills out the ballot type via sampling without replacement from the preference
    interval (i.e. according to the name-Plackett-Luce model).

    Args:
        config (BlocSlateConfig): Configuration object containing all necessary parameters for
            working with a bloc-slate ballot generator.
        group_ballots (bool): If True, group identical ballots in the returned profile and
            set the weight accordingly. Defaults to True.

    Returns:
        RankProfile: Merged ``RankProfile`` object generated by the model.
    """
    config.is_valid(raise_errors=True)

    pp_by_bloc = _inner_slate_plackett_luce(config)

    pp = RankProfile(ballots=tuple())
    for profile in pp_by_bloc.values():
        pp += profile

    if group_ballots:
        pp = pp.group_ballots()

    return pp


def slate_pl_profiles_by_bloc_generator(
    config: BlocSlateConfig,
    *,
    group_ballots: bool = True,
) -> dict[str, RankProfile]:
    """
    Generates a dictionary mapping bloc names to preference profiles using the slate-Plackett-Luce
    model.

    This model first samples a ballot type by flipping a cohesion parameter weighted coin.
    It then fills out the ballot type via sampling without replacement from the preference
    interval (i.e. according to the name-Plackett-Luce model).

    Args:
        config (BlocSlateConfig): Configuration object containing all necessary parameters for
            working with a bloc-slate ballot generator.
        group_ballots (bool): If True, group identical ballots in the returned profile and
            set the weight accordingly. Defaults to True.

    Returns:
        dict[str, RankProfile]: Dictionary whose keys are bloc strings and values are
            ``RankProfile`` objects generated by the model for each bloc.
    """
    config.is_valid(raise_errors=True)

    pp_by_bloc = _inner_slate_plackett_luce(config)

    if group_ballots:
        for bloc, profile in pp_by_bloc.items():
            pp_by_bloc[bloc] = profile.group_ballots()

    return pp_by_bloc



from __future__ import annotations
from ...ballot import ScoreBallot
from typing import Tuple
from .csv_utils import (
    _validate_csv_ballot_weight,
    _validate_csv_ballot_voter_set,
    _validate_csv_ballot_row_break_idxs,
)

CANDIDATES_HEADER_ROW = 1
CANDIDATES_MAPPING_ROW = CANDIDATES_HEADER_ROW + 1
VOTER_SET_HEADER_ROW = 3
VOTER_SET_VALUE_ROW = VOTER_SET_HEADER_ROW + 1
EQ_SIGN_ROW = 5
COLUMN_NAMES_ROW = 6
DATA_START_ROW = COLUMN_NAMES_ROW + 1


def _parse_profile_data_from_score_csv(
    csv_data: list[list[str]],
) -> Tuple[dict[str, str], bool, list[int]]:
    """
    Parse the profile data from a ScoreProfile csv.

    Args:
        csv_data (list[list[str]]): Data from csv.

    Returns:
        Tuple[dict[str, str],  bool, list[int]]:
            inv_candidate_mapping, max_ranking_length,
            includes_voter_set, break_indices
    """
    candidate_row = [c_tuple.strip("()").split(":") for c_tuple in csv_data[2]]
    inv_candidate_mapping = {prefix: cand for cand, prefix in candidate_row}

    includes_voter_set = csv_data[VOTER_SET_VALUE_ROW][0] == "True"

    ballot_data_column_names = csv_data[COLUMN_NAMES_ROW]
    break_indices = [
        i for i, col_name in enumerate(ballot_data_column_names) if col_name == "&"
    ]

    return (
        inv_candidate_mapping,
        includes_voter_set,
        break_indices,
    )


def _parse_ballot_from_score_csv(
    ballot_row: list[str],
    includes_voter_set: bool,
    break_indices: list[int],
    inv_candidate_mapping: dict[str, str],
) -> ScoreBallot:
    """
    Parse a ballot from a ScoreProfile csv row.

    Args:
        ballot_row (list[str]): Row from the csv file containing ballot data.
        includes_voter_set (bool): Whether or not the csv contains voter sets.
        break_indices (list[int]): Where the columns of the csv change from one data type to
            another.
        inv_candidate_mapping (dict[str, str]): The iverted candidate mapping of prefix
            to the cand.

    Returns:
        ScoreBallot: Ballot formatted from row of csv.
    """
    candidates = list(inv_candidate_mapping.values())
    voter_set = set()

    try:
        weight = float(ballot_row[break_indices[0] + 1])
    except Exception:
        raise RuntimeError(
            f"Invalid weight format in ballot row: {ballot_row[break_indices[0] + 1]}. Weight "
            "must be float."
        )

    scores = {
        c: float(ballot_row[i]) for i, c in enumerate(candidates) if ballot_row[i]
    }

    if includes_voter_set:
        voter_set = set(v.strip() for v in ballot_row[break_indices[-1] + 1 :])

    return ScoreBallot(scores=scores, voter_set=voter_set, weight=weight)


def _validate_score_csv_header_values(header_data: list[list[str]]):
    """
    Validate that the values of the header rows are correct.

    Args:
        header_data (list[list[str]]): The rows of the csv file.

    Raises:
        ValueError: If csv header rows are improperly formatted for VoteKit.
    """
    boiler_plate = (
        "This usually indicates that you are loading a csv that was not made with "
        "PreferenceProfile.to_csv()."
    )

    if any(char not in c_tuple for c_tuple in header_data[2] for char in "(:)"):
        raise ValueError(
            (
                f"csv file is improperly formatted. Row {CANDIDATES_MAPPING_ROW} should contain "
                "tuples mapping candidates "
                "to their unique prefixes. For example, (Chris:Ch), (Colleen: Co). "
                f"Not {header_data[2]}. " + boiler_plate
            )
        )

    if len(header_data[VOTER_SET_VALUE_ROW]) != 1 or header_data[VOTER_SET_VALUE_ROW][
        0
    ] not in ["True", "False"]:
        raise ValueError(
            (
                f"csv file is improperly formatted. Row {VOTER_SET_VALUE_ROW} should be 'True' or "
                f"'False', not {header_data[VOTER_SET_VALUE_ROW]}. " + boiler_plate
            )
        )


def _validate_score_csv_header_rows(header_data: list[list[str]]):
    """
    Validate that the names of the header rows are correct.

    Args:
        header_data (list[list[str]]): The rows of the csv file.

    Raises:
        ValueError: If csv headers rows are improperly formatted for VoteKit.
    """
    boiler_plate = (
        "This usually indicates that you are loading a csv that was not made with "
        "PreferenceProfile.to_csv()."
    )

    if header_data[0] != ["VoteKit ScoreProfile"]:
        raise ValueError(
            (
                "csv file is improperly formatted. Row 0 should be 'VoteKit ScoreProfile',"
                f"not {header_data[0]}. " + boiler_plate
            )
        )

    if header_data[CANDIDATES_HEADER_ROW] != ["Candidates"]:
        raise ValueError(
            (
                f"csv file is improperly formatted. Row {CANDIDATES_HEADER_ROW} should be "
                f"'Candidates', not {header_data[CANDIDATES_HEADER_ROW]}. "
                + boiler_plate
            )
        )

    if header_data[VOTER_SET_HEADER_ROW] != ["Includes Voter Set"]:
        raise ValueError(
            (
                f"csv file is improperly formatted. Row {VOTER_SET_HEADER_ROW} should be"
                f" 'Includes Voter Set', not {header_data[VOTER_SET_HEADER_ROW]}. "
                + boiler_plate
            )
        )

    if header_data[EQ_SIGN_ROW] != ["="] * 10:
        raise ValueError(
            (
                f"csv file is improperly formatted. Row {EQ_SIGN_ROW} should be "
                "'=,=,=,=,=,=,=,=,=,=', "
                f"not {header_data[EQ_SIGN_ROW]}. " + boiler_plate
            )
        )


def _validate_score_csv_header(header_data: list[list[str]]):
    """
    Validate that the values of the header rows are correct.

    Args:
        header_data (list[list[str]]): The rows of the header of the csv file.

    Raises:
        ValueError: If csv header is improperly formatted for VoteKit.
    """

    _validate_score_csv_header_rows(header_data)
    _validate_score_csv_header_values(header_data)


def _validate_score_csv_ballot_header_row(
    ballot_rows: list[list[str]],
    candidate_prefixes: tuple[str],
    include_voter_set: bool,
):
    """
    Validate that the ballot header row is formatted correctly.

    Args:
        ballot_rows (list[list[str]]): All ballot rows.
        candidate_prefixes (tuple[str]): The tuple of candidate prefixes.
        max_ranking_length (int): The max ranking length.
        include_voter_set (bool): Whether or not there is a voter set.

    Raises:
        ValueError: If the ballot header is improperly formatted for VoteKit.
    """
    boiler_plate = (
        "This usually indicates that you are loading a csv that was not made with "
        "PreferenceProfile.to_csv()."
    )

    header_row = ballot_rows[0]

    break_indices = [i for i, col_name in enumerate(header_row) if col_name == "&"]

    # TODO what is this if condition?
    if len(header_row[: break_indices[0]]) > 0 and any(
        c not in header_row for c in candidate_prefixes
    ):
        raise ValueError(
            (
                f"csv file is improperly formatted. Row {COLUMN_NAMES_ROW} should include all "
                "candidates before the first &. " + boiler_plate
            )
        )

    if "Weight" not in header_row:
        raise ValueError(
            (
                f"csv file is improperly formatted. Row {COLUMN_NAMES_ROW} should include"
                " 'Weight' column. " + boiler_plate
            )
        )

    if (include_voter_set and "Voter Set" not in header_row) or (
        not include_voter_set and "Voter Set" in header_row
    ):
        raise ValueError(
            (
                "csv file is improperly formatted. Includes Voter Set is not set to the correct"
                f" value given the columns in row {COLUMN_NAMES_ROW}: {header_row}. "
                + boiler_plate
            )
        )


def _validate_score_csv_ballot_score(
    ballot_row: list[str],
    row_index: int,
    candidates: tuple[str],
):
    """
    Validate that the ballot scores are formatted correctly.

    Args:
        ballot_row (list[str]): A ballot row.
        row_index (int): The index of the row in the csv, 0-indexed.
        candidates (tuple[str]): The tuple of candidates.

    Raises:
        ValueError: If the ballot scores are improperly formatted for VoteKit.
    """
    boiler_plate = (
        "This usually indicates that you are loading a csv that was not made with "
        "PreferenceProfile.to_csv()."
    )

    break_idx = ballot_row.index("&")

    if len(ballot_row[:break_idx]) != len(candidates):
        raise ValueError(
            (
                f"csv file is improperly formatted. Ballot in row {row_index} is missing "
                f"some scores: {ballot_row[:break_idx]} ." + boiler_plate
            )
        )

    for score in ballot_row[:break_idx]:
        if score == "":
            continue

        try:
            float(score)
        except ValueError:
            raise ValueError(
                (
                    f"csv file is improperly formatted. Ballot in row {row_index} has "
                    f"non-float score value: {score}. " + boiler_plate
                )
            )


def _validate_score_csv_ballot_rows(csv_data: list[list[str]]):
    """
    Validate that each ballot row is formatted correctly.

    Args:
        csv_data (list[list[str]]): The full csv.

    Raises:
        ValueError: If a row of the csv is improperly formatted for VoteKit.
    """
    candidate_row = csv_data[CANDIDATES_MAPPING_ROW]
    include_voter_set_row = csv_data[VOTER_SET_VALUE_ROW]

    candidate_tuples = [c_tuple.strip("()").split(":") for c_tuple in candidate_row]
    candidates, candidate_prefixes = zip(*candidate_tuples)

    include_voter_set = include_voter_set_row[0] == "True"

    _validate_score_csv_ballot_header_row(
        csv_data[COLUMN_NAMES_ROW:], candidate_prefixes, include_voter_set
    )

    error_rows = []

    for i, ballot_row in enumerate(csv_data[DATA_START_ROW:]):
        try:
            _validate_csv_ballot_row_break_idxs(ballot_row, i + DATA_START_ROW)
            _validate_score_csv_ballot_score(
                ballot_row,
                i + DATA_START_ROW,
                candidates,
            )
            _validate_csv_ballot_weight(ballot_row, i + DATA_START_ROW)
            _validate_csv_ballot_voter_set(
                ballot_row, i + DATA_START_ROW, include_voter_set
            )
        except ValueError as e:
            error_rows.append((i + DATA_START_ROW, e))

        if len(error_rows) == 10:
            raise ValueError(
                (
                    f"There are errors on at least ten lines of this csv: {error_rows}."
                    " Did you try to load a csv that was edited by hand and not "
                    "created with the PreferenceProfile.to_csv() method?"
                )
            )
    if len(error_rows) > 0:
        raise ValueError(
            (
                f"There are errors on {len(error_rows)} lines of this csv: {error_rows}."
                " Did you try to load a csv that was edited by hand and not "
                "created with the PreferenceProfile.to_csv() method?"
            )
        )


def _validate_score_csv_format(csv_data: list[list[str]]):
    """
    Validate that the csv is properly formatted for VoteKit.

    Args:
        csv_data (list[list[str]]): The rows of the csv file.

    Raises:
        ValueError: If csv is improperly formatted for VoteKit.
    """
    _validate_score_csv_header(csv_data[:COLUMN_NAMES_ROW])
    _validate_score_csv_ballot_rows(csv_data)



"""
Generate ranked preference profiles using the Impartial Anonymous Culture (IAC) model.

The main API functions in this module are:

- `iac_profile_generator`: Generates a single preference profile using the IAC distribution.
"""

import math
import numpy as np
import random
from collections import Counter
from functools import lru_cache
from typing import Optional, Sequence

from votekit.pref_profile import RankProfile
from votekit.utils import index_to_lexicographic_ballot, build_df_from_ballot_samples

# ====================================================
# ================= Helper Functions =================
# ====================================================


@lru_cache
def _total_num_ballots(n_candidates: int, max_ballot_length: int) -> int:
    """
    Calculate the total number of possible ballots given the number of candidates and the maximum
    ballot length.

    Args:
        n_candidates (int): Number of candidates.
        max_ballot_length (int): Maximum length of each ballot.

    Returns:
        int: Total number of possible ballots.
    """
    return sum(
        math.comb(n_candidates, i) * math.factorial(i)
        for i in range(1, max_ballot_length + 1)
    )


# =================================================
# ================= API Functions =================
# =================================================


def iac_profile_generator(
    candidates: Sequence[str],
    number_of_ballots: int,
    max_ballot_length: Optional[int] = None,
) -> RankProfile:
    """
    Generate a profile according to the Impartial Anonymous Culture (IAC) model where each profile
    is equally likely.

    Args:
        candidates (Sequence[str]): List of candidate strings.
        number_of_ballots (int): Number of ballots to generate.
        max_ballot_length (Optional[int]): Maximum length of each ballot. If None, defaults to
            the number of candidates.

    Returns:
        RankProfile: Generated rank profile
    """

    if max_ballot_length is None:
        max_ballot_length = len(candidates)

    # choose index as sampled 0 to N, do this n! times
    num_cands = len(candidates)
    num_gaps = number_of_ballots  # + 1
    gap_freq = np.zeros(num_gaps, dtype=int)  # record the number of gaps in stars/bars

    # rather than iterate n! times, we perform multiple
    # multinomial experiments
    num_valid_ballots = _total_num_ballots(num_cands, max_ballot_length)
    for _ in range(num_valid_ballots):
        gap_freq[random.randint(0, num_gaps) - 1] += 1

    # TODO: Double check possible off by 1 errors here and in `gap_freq`
    ballot_indices = np.cumsum(gap_freq) - 1
    ballots_as_cand_ind = [
        tuple(
            index_to_lexicographic_ballot(
                ballot_ind,
                num_cands,
                max_ballot_length,
                _total_num_ballots,
                always_use_total_valid_ballots_method=False,
            )
        )
        for ballot_ind in ballot_indices
    ]
    ballots_as_counter = Counter(ballots_as_cand_ind)
    pp_df = build_df_from_ballot_samples(dict(ballots_as_counter), candidates)
    pp_df.index.name = "Ballot Index"
    return RankProfile(
        df=pp_df,
        max_ranking_length=max_ballot_length,
        candidates=candidates,
    )



from __future__ import annotations


def _validate_csv_ballot_weight(
    ballot_row: list[str],
    row_index: int,
):
    """
    Validate that the ballot weight is formatted correctly.

    Args:
        ballot_row (list[str]): A ballot row.
        row_index (int): The index of the row in the csv, 0-indexed.

    Raises:
        ValueError: If the ballot weight is improperly formatted for VoteKit.
    """
    boiler_plate = (
        "This usually indicates that you are loading a csv that was not made with "
        "PreferenceProfile.to_csv()."
    )

    break_idxs = [i for i, string in enumerate(ballot_row) if string == "&"]
    if break_idxs[1] - break_idxs[0] != 2:
        raise ValueError(
            (
                f"csv file is improperly formatted. Ballot in row {row_index} has a weight"
                " entry that is too long or short. " + boiler_plate
            )
        )

    else:

        try:
            float(ballot_row[break_idxs[0] + 1])

        except ValueError:
            raise ValueError(
                (
                    f"csv file is improperly formatted. Ballot in row {row_index} has a "
                    "weight entry that can't be converted to float "
                    f"{ballot_row[break_idxs[0] + 1]}. " + boiler_plate
                )
            )


def _validate_csv_ballot_voter_set(
    ballot_row: list[str], row_index: int, include_voter_set: bool
):
    """
    Validate that the ballot voter set is formatted correctly.

    Args:
        ballot_row (list[str]): A ballot row.
        row_index (int): The index of the row in the csv, 0-indexed.
        include_voter_set (bool): Whether or not there is a voter set.

    Raises:
        ValueError: If the ballot voter set is improperly formatted for VoteKit.
    """

    boiler_plate = (
        "This usually indicates that you are loading a csv that was not made with "
        "PreferenceProfile.to_csv()."
    )

    break_idxs = [i for i, string in enumerate(ballot_row) if string == "&"]

    if not include_voter_set and len(ballot_row[break_idxs[-1] + 1 :]) > 0:
        raise ValueError(
            (
                f"csv file is improperly formatted. Ballot in row {row_index} has a "
                f"voter set but it should not: {ballot_row[break_idxs[-1] + 1 :]} "
                + boiler_plate
            )
        )


def _validate_csv_ballot_row_break_idxs(ballot_row: list[str], row_index: int):
    """
    Validate that the ballot number of & symbols in each row is correct.

    Args:
        ballot_row (list[str]): A ballot row.
        row_index (int): The index of the row in the csv, 0-indexed.

    Raises:
        ValueError: If the ballot row is improperly formatted for VoteKit.
    """
    boiler_plate = (
        "This usually indicates that you are loading a csv that was not made with "
        "PreferenceProfile.to_csv()."
    )
    break_idxs = [i for i, string in enumerate(ballot_row) if string == "&"]

    if len(break_idxs) != 2:
        raise ValueError(
            (
                f"csv file is improperly formatted. Ballot in row {row_index} does not have 2 &"
                f"symbols, it has {len(break_idxs)}." + boiler_plate
            )
        )



from .abstract_ranking import RankingElection
from votekit.pref_profile import RankProfile
from votekit.elections.election_state import ElectionState
from votekit.cleaning import remove_and_condense_rank_profile
from votekit.utils import first_place_votes
from votekit.elections.election_types.ranking import Plurality, STV
from votekit.elections.transfers import fractional_transfer
from votekit.ballot import RankBallot
from typing import Optional, Callable, Union, Literal
from functools import partial


class Alaska(RankingElection):
    """
    Alaska election. Election method that first runs a Plurality election to choose a user-specified
    number of final-round candidates, then runs STV to choose :math:`m` winners.

    Args:
        profile (RankProfile): Profile to conduct election on.
        m_1 (int, optional): Number of final round candidates, i.e. number of winners of Plurality
            round. Defaults to 2.
        m_2 (int, optional): Number of seats to elect in STV round, i.e. number of overall winners.
            Defaults to 1.
        transfer (Callable[[str, float], Union[tuple[Ballot], list[Ballot]], int],
            tuple[Ballot,...]], optional):
            Transfer method. Defaults to fractional transfer.
            Function signature is elected candidate, their number of first-place votes, the list of
            ballots with them ranked first, and the threshold value. Returns the list of ballots
            after transfer.
        quota (str, optional): Formula to calculate quota. Accepts "droop" or "hare".
            Defaults to "droop".
        simultaneous (bool, optional): True if all candidates who cross threshold in a round are
            elected simultaneously, False if only the candidate with highest first-place votes
            who crosses the threshold is elected in a round. Defaults to True.
        tiebreak (str, optional): Tiebreak method to use. Options are None, 'random', and 'borda'.
            Defaults to None, in which case a tie raises a ValueError.
        fpv_tie_convention (Literal["high", "average", "low"], optional): How to award points
            for tied first place votes. Defaults to "average", where if n candidates are tied for
            first, each receives 1/n points. "high" would award them each one point, and "low" 0.
            Only used by ``score_function`` parameter.

    """

    def __init__(
        self,
        profile: RankProfile,
        m_1: int = 2,
        m_2: int = 1,
        transfer: Callable[
            [str, float, Union[tuple[RankBallot], list[RankBallot]], int],
            tuple[RankBallot, ...],
        ] = fractional_transfer,
        quota: str = "droop",
        simultaneous: bool = True,
        tiebreak: Optional[str] = None,
        fpv_tie_convention: Literal["high", "low", "average"] = "average",
    ):
        if m_1 <= 0:
            raise ValueError("m_1 must be positive.")
        elif m_2 <= 0:
            raise ValueError("m_2 must be positive.")
        elif m_1 < m_2:
            raise ValueError("m_1 must be greater than or equal to m_2.")
        self.m_1 = m_1
        self.m_2 = m_2
        self.transfer = transfer
        self.quota = quota
        self.simultaneous = simultaneous
        self.tiebreak = tiebreak
        super().__init__(
            profile,
            score_function=partial(
                first_place_votes, tie_convention=fpv_tie_convention
            ),
            sort_high_low=True,
        )

    def get_profile(self, round_number: int = -1) -> RankProfile:
        """
        Fetch the RankProfile of the given round number.

        Args:
            round_number (int, optional): The round number. Supports negative indexing. Defaults to
                -1, which accesses the final profile.

        Returns:
            RankProfile

        """
        if (
            round_number < -len(self.election_states)
            or round_number > len(self.election_states) - 1
        ):
            raise IndexError("round_number out of range.")

        round_number = round_number % len(self.election_states)

        profile = self._profile

        if round_number in [0, 1]:
            for i in range(round_number):
                profile = self._run_step(profile, self.election_states[i])

        else:
            stv = STV(
                self.get_profile(1),  # plurality profile
                self.m_2,
                self.transfer,
                self.quota,
                self.simultaneous,
                self.tiebreak,
            )

            # e.g., round 2 of Alaska is equivalent to round 1 of stv
            profile = stv.get_profile(round_number - 1)

        return profile

    def _is_finished(self):
        elected_cands = [c for s in self.get_elected() for c in s]

        if len(elected_cands) == self.m_2:
            return True
        return False

    def _run_step(
        self, profile: RankProfile, prev_state: ElectionState, store_states=False
    ) -> RankProfile:
        """
        Run one step of an election from the given profile and previous state.

        Args:
            profile (RankProfile): Profile of ballots.
            prev_state (ElectionState): The previous ElectionState.
            store_states (bool, optional): True if `self.election_states` should be updated with the
                ElectionState generated by this round. This should only be True when used by
                `self._run_election()`. Defaults to False.

        Returns:
            RankProfile: The profile of ballots after the round is completed.
        """
        if prev_state.round_number == 0:
            plurality = Plurality(profile, self.m_1, self.tiebreak)
            remaining = plurality.get_elected()
            eliminated = plurality.get_remaining()
            tiebreaks = plurality.election_states[-1].tiebreaks
            new_profile: RankProfile = remove_and_condense_rank_profile(
                [c for s in eliminated for c in s],
                profile,
            )

            if self.score_function is None:
                raise ValueError("score_function must be defined for Alaska election.")
            scores = self.score_function(new_profile)

            if store_states:
                new_state = ElectionState(
                    round_number=prev_state.round_number + 1,
                    remaining=remaining,
                    eliminated=eliminated,
                    scores=scores,
                    tiebreaks=tiebreaks,
                )

                self.election_states.append(new_state)

        else:
            stv = STV(
                profile,
                self.m_2,
                self.transfer,
                self.quota,
                self.simultaneous,
                self.tiebreak,
            )
            new_profile = stv.get_profile()

            if store_states:
                # first state was already stored by Plurality
                for state in stv.election_states[1:]:
                    state.round_number += 1
                self.election_states += stv.election_states[1:]

        return new_profile
